#include "Planner.h"

#include "Delegator.h"
#include "DeviceType.h"
#include "Errors.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <iostream>
#include <thread>
#include <chrono>

using namespace std;
using namespace std::chrono;

/* Public methods: */

Planner::Planner(std::string uartPort, std::string gpioNum) {
	pass = 0;
	lockR = false;
	lockW = false;
	lockRW = false;
	lockDR = false;
	updateW = false;
	updateRW = false;
	halted = false;
	confirmed = false;

	this->uartPort = uartPort;
	this->gpioNum = gpioNum;
}

Planner::~Planner() {
}

void Planner::setDelegator(Delegator *dlg) {
	this->pDelegator = dlg;
}

void Planner::addNode(Node &node) {
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::R) {
		nowNodesR.insert(make_pair(node.getAddress(), node));
	}
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::RW) {
		nowNodesRW.insert(make_pair(node.getAddress(), node));
	}
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::W) {
		nowNodesW.insert(make_pair(node.getAddress(), node));
	}
	if (node.getType().getLatency() == DeviceType::DELAYED
			&& node.getType().getReadWrite() == DeviceType::R) {
		delayedNodesR.insert(make_pair(node.getAddress(), node));
	}
}

void Planner::addType(DeviceType &type) {
	types.insert(make_pair(type.getType(), type));
}

void Planner::removeNode(Node &node) {
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::W) {
		while (true) {
			if (lockW == false) {
				nowNodesW.erase(node.getAddress());
				break;
			}
		}
	}
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::R) {
		while (true) {
			if (lockR == false) {
				nowNodesR.erase(node.getAddress());
				break;
			}
		}
	}
	if (node.getType().getLatency() == DeviceType::NOW
			&& node.getType().getReadWrite() == DeviceType::RW) {
		while (true) {
			if (lockRW == false) {
				nowNodesRW.erase(node.getAddress());
				break;
			}
		}
	}
	if (node.getType().getLatency() == DeviceType::DELAYED
			&& node.getType().getReadWrite() == DeviceType::R) {
		while (true) {
			if (lockDR == false) {
				delayedNodesR.erase(node.getAddress());
				break;
			}
		}
	}

}

void Planner::setNodeData(Node &node) {
	if (node.getType().getReadWrite() == DeviceType::W) {
		while (true) {
			if (lockW == false) {
				updateW = true;
				nowNodesW[node.getAddress()] = node;
				updateW = false;
				break;
			}
		}
	}
	if (node.getType().getReadWrite() == DeviceType::RW) {
		while (true) {
			if (lockRW == false) {
				updateRW = true;
				nowNodesRW[node.getAddress()] = node;
				updateRW = false;
				break;
			}
		}
	}

}

void Planner::setAddress(unsigned char address) {
	this->address = address;
	message.setAddress(address);
}

void Planner::start() {
	gpio4 = new GPIO(gpioNum);
	gpio4->setdirGpio("out");

	int err = uart.openUART(uartPort, 115200);
	if (err == 1) {
		pDelegator->logError(ERROR_UART);
	} else {
		std::thread loopT([=] {loop();});
		loopT.detach();
	}
}

/* Private methods: */

int Planner::fetchData(Node &node, int length) {
	int rxpos = 0;
	int delta = 0;

	bool start = false;
	bool end = false;

	while (rxpos < length && delta < 80) {
		usleep(1000);
		n = uart.pollComport(rxbufpart, 100);
		if (n > 0) {
			for (int j = 0; j < n; j++) {
				rxbuf[rxpos] = rxbufpart[j];
				rxpos++;
				if (rxbufpart[j] == START_CHAR) {start = true;}
				if (rxbufpart[j] == END_CHAR) {end = true;}
			}
		}
		delta++;
	}

	uart.flush();
	//pokud je paket nekompletni, vypis do konzole
	if (start == false || end == false) cout << "-------------" << endl;
	if (start == false) {
		cout << "Missing start from node address " << node.getAddress() << endl;
	}
	if (rxpos < length) {
		cout << "Received " << rxpos << " bytes instead of " << length << endl;
	}
	if (end == false) {
		cout << "Missing end from node address " << node.getAddress() << endl;
	}
	if (start == false || end == false) cout << "-------------" << endl;
	return rxpos;
}

void Planner::requestGetNodes(nodeMap& nodes) {
	for (nodeMap::iterator iter = nodes.begin(); iter != nodes.end(); iter++) {
		Node node = iter->second;
		//posleme request
		gpio4->setvalGpio("1");
		usleep(5);
		n = message.makeRequest(txbuf, node);
		uart.sendBuf(txbuf, n);
		gpio4->setvalGpio("0");

		int rxpos = fetchData(node, node.getDataLength() + 8);

		int change;
		change = message.get(rxbuf, node, rxpos);
		if (change == 1) {
			pDelegator->updateNodeData(node);
		}

		if (change == -1) {
			milliseconds errorTime = duration_cast<milliseconds>(
					high_resolution_clock::now().time_since_epoch());

			unsigned long currentErrorTime = errorTime.count();

			if (currentErrorTime - node.getErrorTime() < 20000) {
				pDelegator->logError(ERROR_PACKET_NONSENSE, node.getAddress());
			}
			node.setErrorTime(currentErrorTime);
		}

		iter->second = node;

	}
}

void Planner::setNodes(nodeMap& nodes) {
	gpio4->setvalGpio("1");
	usleep(1000);
	for (nodeMap::iterator iter = nodes.begin(); iter != nodes.end(); iter++) {
		Node node = iter->second;
		n = message.makeSet(txbuf, node);
		uart.sendBuf(txbuf, n);
		usleep(500);
	}
	gpio4->setvalGpio("0");
	usleep(1000);
}

void Planner::meassureNodes(nodeMap& nodes) {
	gpio4->setvalGpio("1");
	usleep(1000);
	for (nodeMap::iterator iter = nodes.begin(); iter != nodes.end(); iter++) {
		Node node = iter->second;
		n = message.makeMeassure(txbuf, node);
		uart.sendBuf(txbuf, n);
		usleep(500);
	}
	gpio4->setvalGpio("0");
	usleep(1000);
}

void Planner::loop() {
	while (true) {
		if (!halted) {
			confirmed = false;
			pass++;

			lockR = true;
			requestGetNodes(nowNodesR);
			lockR = false;
			usleep(1000);

			if (updateW == false) {
				lockW = true;
				setNodes(nowNodesW);
				lockW = false;
			}

			if (updateRW == false) {
				lockRW = true;
				setNodes(nowNodesRW);
				lockRW = false;
			}

			if (pass == 10 && updateRW == false) {
				lockRW = true;
				requestGetNodes(nowNodesRW);
				lockRW = false;
			}

			if (pass == 50) {
				lockDR = true;
				meassureNodes(delayedNodesR);
				lockDR = false;
			}

			if (pass == 100) {
				lockDR = true;
				requestGetNodes(delayedNodesR);
				lockDR = false;
				pass = 0;
			}
			confirmed = true;
		} else {

		}
		usleep(1000);
	}
}

void Planner::discoverNewNodes() {
	//cekame az planovac dokonci cyklus a zastavime ho
	while (!halted) {
		if (confirmed) {
			halted = true;
		}
		usleep(100);
	}

	newNodesMap.clear();

	//poslat discovery packet novym nodum
	gpio4->setvalGpio("1");
	usleep(2000);
	n = message.makeDiscovery(txbuf);
	uart.sendBuf(txbuf, n);
	usleep(800);
	gpio4->setvalGpio("0");

	//data polling
	int rxpos = 0;
	for (int i = 0; i < 8000; i++) {
		usleep(1000);
		n = uart.pollComport(rxbufpart, 100);
		if (n > 0) {
			cout << n << endl;
			for (int j = 0; j < n; j++) {
				rxbuf[rxpos] = rxbufpart[j];
				rxpos++;
			}
		}
	}

	if (rxpos == 0) {
		pDelegator->logError(ERROR_NO_NEW_NODES);
		halted = false;
	}

	if ((rxpos >= 9)) {
		message.discoveryResponse(rxbuf, rxpos, newNodesMap, types);
		if (newNodesMap.size() > 0) {
			usleep(10000);
			pDelegator->requestNewAddresses(newNodesMap);
		}
		if (rxpos % 9 > 0) { //paket je dlouhy 9 bajtu, pokud tedy prijaty pocet neni delitelny deviti, mame objeveny ne vsechny nody
			pDelegator->logError(ERROR_PACKET_NONSENSE);
			halted = false;
		}
	} else if (rxpos > 1 && rxpos < 9) {
		pDelegator->logError(ERROR_PACKET_NONSENSE);
		halted = false;
	}

}

void Planner::addressNewNodes(nodeMap &newNodes) {
	//poslat discovery address packet novym nodum
	for (nodeMap::iterator iter = newNodes.begin(); iter != newNodes.end();
			iter++) {
		Node node = iter->second;

		for (int c = 1; c < 20; c++) {
			bool parse = false;

			gpio4->setvalGpio("1");
			usleep(2000);
			n = message.makeDiscoveryAddress(txbuf, node);
			uart.sendBuf(txbuf, n);
			gpio4->setvalGpio("0");

			//prijem dat
			int rxpos = fetchData(node, 10);

			parse = message.discoveryCheck(rxbuf, rxpos, node);

			cout << "Adresace nodu " << c << " "<< node.getAddress() << " random " << node.getRandom() << " prijado " << rxpos << " parsovano: " << parse << endl;

			if (parse) {
				gpio4->setvalGpio("1");
				usleep(2000);
				for (int i = 0; i < 5; i++) {
					n = message.makeDiscoveryCheckOk(txbuf, node);
					uart.sendBuf(txbuf, n);
					usleep(5000);
				}
				gpio4->setvalGpio("0");
				addNode(node);
				pDelegator->addNodeToMap(node);
				usleep(100000);
				break;
			}

			usleep(10000);
		}
	}
	halted = false;
}

